#!/usr/bin/perl -w
# wogri@wogri.com

use strict; 
use Net::LDAP; 
use Data::Dumper;
use Getopt::Std; 
use POSIX qw(strftime);

STDOUT->autoflush(1);

my $passwd="yourpasswd";
my $host="your.ldap.server.your.domain";
my $user="cn=daemon,dc=your,dc=domain";
my $base="ou=dnsConfig,dc=your,dc=domain";
my $path="/var/named/etc/masters/";
my $sudo=""; # blank for no sudo.
my $reload_dns="$sudo rndc reload "; 
my %options=(); 
getopts("hH:p:u:b:P:v",\%options) or &usage; 
&usage if $options{h};

my @domainRecords=('aRecord','mDRecord','mXRecord','nSRecord','cNAMERecord','TXTRecord','SRVRecord','PTRRecord','AAAARecord','SPFRecord','RAWRecord'); 
my %domainRecordMapping=(
	aRecord=>'A', 
	mDRecord=>'MD', 
	mXRecord=>'MX', 
	nSRecord=>'NS', 
	cNAMERecord=>'CNAME', 
	TXTRecord=>'TXT',
	SRVRecord=>'SRV',
	PTRRecord=>'PTR',
	AAAARecord=>'AAAA'
	); 

$passwd=$options{p} if not defined $passwd;
$host=$options{h} if not defined $host;
$user=$options{u} if not defined $user;
$base=$options{b} if not defined $base;
$path=$options{P} if not defined $path;

my $ldap = Net::LDAP->new($host, async=>0, onerror=>'warn', version=>'3') or die "dead: $@";
$ldap->start_tls(verify=>'none') or die "dead: $@"; 
my $mesg = $ldap->bind($user, password=>$passwd, version=>3) or die "$@";
my %domains; 
&buildTree($base,\%domains,$ldap,"",undef);
&writeData(\%domains,$path); 

sub usage {
	print "usage: $0 [options]
where options can be:
-p	LDAP-Password
-H 	LDAP-Host (URI-Support, e.g. ldaps://your.ldap.host:1234)
-u 	LDAP-Username (e.g. cn=yourUser,dc=your,dc=domain)
-b	LDAP-Search-Base (e.g. ou=DNSConfig,dc=your,dc=domain)
-P  Filesystem-Path where to write the finished config.
-h  this help

All these options can be set in $0 (head section), so you
don't need to type them in your command-line (which is
not a good idea to do, other people would see the password)

This software is licensed under the GPLv2, see www.gpl.org
";
  exit 0;
}

sub buildTree {
	my $dom=shift;
	my $domainsRef=shift;
	my $ldap=shift;
	my $fqdn=shift;
	my $domain=shift;
	print "investigating: $dom\n" if $options{v}; 
	my $mesg=$ldap->search( base=>$dom, scope=>'one', filter=>'(&(objectclass=dNSDomain))');
	$mesg->code && die "dieing because of: ".$mesg->error;
	my $entries=$mesg->entries; 
	if ($entries>0) {
		foreach my $entry ($mesg->entries) { 
			my $soa=&getAttribute("sOARecord",$entry); 
			my $dc=&getAttribute("dc",$entry); 
			my $fqdn="$$dc[0].$fqdn";
			if (defined $soa) {
				if (@$soa > 1) {
					print STDERR "Error: Multiple SOA Records found in entry: $$entry{asn}{objectName}: Aborting...\n";
					exit 1; 
				}
				$$domainsRef{$fqdn}{soa}=$$soa[0]; 
				$domain=$fqdn;
			}
			foreach (@domainRecords) {
				my $val=&getAttribute($_, $entry); 
				if (defined $val) {
					$$domainsRef{$domain}{$_}{$fqdn}=$val; 
				}
			}
    	&buildTree($$entry{asn}{objectName},$domainsRef,$ldap,$fqdn,$domain); 
		}
	}
}

sub getAttribute {
	my $attrib=shift;
	my $entryRef=shift;
	my @elems=@{$$entryRef{asn}{attributes}};
	foreach my $e (@elems) {
		if ($$e{type} eq $attrib) {
			return $$e{vals};
		}
	}
	return undef;
}

sub writeData {
	my $domainsRef=shift;
	my $path=shift;
	foreach my $domain (keys %$domainsRef) {
		my $filename="$path/$domain"."db";
		open (FH,"> $filename.bak"); 
		my ($primaryDNS,$mailAddr,$serial,$refresh,$retry,$expire,$ttl)=split(/\s/,$$domainsRef{$domain}{soa});
		$serial=strftime "%Y%m%d%H", localtime; # this is not a good way to do this, the serial should be read and not updated like I do it...
		print FH "$domain $ttl IN SOA $primaryDNS $mailAddr $serial $refresh $retry $expire $ttl\n"; 
		foreach my $recordType (@domainRecords) {
			print FH "\n; ${recordType}s: \n";
			foreach my $record (keys %{$$domainsRef{$domain}{$recordType}}) {
				foreach my $entry (@{$$domainsRef{$domain}{$recordType}{$record}}) {
					my $tabs="\t\t\t\t\t";
					my $length=length($record); 
					$tabs="\t\t\t\t" if ($length<16);
					$tabs="\t\t\t" if ($length<24 and $length >=16);
					$tabs="\t\t" if ($length<32 and $length>=24);
					$tabs="\t" if ($length>=32);
					print FH "$record${tabs}IN $domainRecordMapping{$recordType}\t$entry\n";
				}
			}
		}
		close (FH);
		# some shell and sed magic
		system 'sed \'s/^\(.*IN SOA.*\)[0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9]\(.*\)$/\1\2/\' '.$filename.' | sort > '.$filename.".bak.ori"; 
		my $diff=system 'sed \'s/^\(.*IN SOA.*\)[0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9][0-9]\(.*\)$/\1\2/\' '.$filename.".bak | sort | diff -u $filename.bak.ori -"; 
		if ($diff) {
			print "reloading DNS zone $domain\n"; 
			chop($domain); 
			system "mv $filename.bak $filename && $reload_dns $domain && rm $filename.bak.ori && $sudo tail -5 /var/log/daemon"; 
		}
		else {
			unlink("$filename.bak.ori"); 
			unlink("$filename.bak"); 
		}
	}
}
$mesg = $ldap->unbind;   # take down session
